﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;

namespace ININ.Supervisor.Chart
{
    internal class AutoQueueEnumerator<T> : IEnumerator<T>
    {
        private readonly T[] _Queue = null;
        private readonly int _Count = 0;
        private readonly int _StartIndex = -1;
        private int _Index = -1;
        private object _Lock;

        public AutoQueueEnumerator(T[] queue, int startIndex, int count, object Lock)
        {
            _Queue = queue;
            _StartIndex = _Index = startIndex;
            _Count = count;
            _Lock = Lock;
        }

        #region IEnumerator<T> Members

        public T Current
        {
            get
            {
                lock (_Lock)
                {
                    if (_Index > -1)
                    {
                        long curIndex = _Index % _Count;
                        return _Queue[curIndex];
                    }
                    throw new ArgumentOutOfRangeException("Invalid index");
                }
            }
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            // nothing to do here
        }

        #endregion

        #region IEnumerator Members

        object IEnumerator.Current
        {
            get { lock (_Lock) { return this.Current; } }
        }

        public bool MoveNext()
        {
            lock (_Lock)
            {
                _Index++;
                return _Count > 0 && _Index - _Count <= _StartIndex;
            }
        }

        public void Reset()
        {
            lock (_Lock)
            {
                _Index = _StartIndex;
            }
        }

        #endregion
    }

    public class AutoQueue<T> : IEnumerable<T>, ICollection, IEnumerable, INotifyCollectionChanged, INotifyPropertyChanged
    {
        readonly private T[] _Queue = null;
        readonly int _Count = 0;
        private int _InsertIndex = 0;
        private object _Lock = new object();

        public AutoQueue(int count)
        {
            _Count = count;
            _Queue = new T[count];
        }

        public int FillLevel
        {
            get { lock (_Lock) { return Math.Min(_InsertIndex, _Count); } }
        }

        public void Enqueue(T value)
        {
            lock (_Lock)
            {
                _Queue[_InsertIndex % _Count] = value;
                ++_InsertIndex;
                NotifyCollectionChangedEventHandler collectionChanged = CollectionChanged;
                if (collectionChanged != null)
                    collectionChanged(this, new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
                PropertyChangedEventHandler propertyChanged = PropertyChanged;
                if (propertyChanged != null)
                    propertyChanged(this, new PropertyChangedEventArgs("Count"));
            }
        }

        #region ICollection Members

        public void CopyTo(Array array, int index)
        {
            throw new NotImplementedException();
        }

        public int Count
        {
            get { return _Count; }
        }

        public bool IsSynchronized
        {
            get { return true; }
        }

        public object SyncRoot
        {
            get { return _Lock; }
        }

        #endregion

        #region IEnumerable Members

        public IEnumerator GetEnumerator()
        {
            lock (_Lock)
            {
                return new AutoQueueEnumerator<T>(_Queue, _InsertIndex - 1, FillLevel, _Lock);
            }
        }

        #endregion

        #region IEnumerable<T> Members

        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            lock (_Lock)
            {
                return new AutoQueueEnumerator<T>(_Queue, _InsertIndex - 1, FillLevel, _Lock);
            }
        }

        #endregion

        #region INotifyCollectionChanged Members

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion

        #region INotifyPropertyChanged Members

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }
}
